#!/usr/bin/python3
#
# This is not a stable API; for use within autopkgtest only.
#
# Copyright 2024 Canonical Ltd.
# Copyright 2024 Simon McVittie
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import shutil
import sys
import subprocess
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Sequence


class Dependency(ABC):
    @abstractmethod
    def check(self) -> bool:
        raise NotImplementedError


class Executable(Dependency):
    def __init__(
        self,
        exe: str,
        package: str,
        *,
        fatal: bool = True,
        if_not_root: bool = False,
        if_not_systemd: bool = False,
    ) -> None:
        self.exe = exe
        self.fatal = fatal
        self.package = package
        self.if_not_root = if_not_root
        self.if_not_systemd = if_not_systemd

    def check(self) -> bool:
        qualifiers: List[str] = []

        if self.if_not_root:
            if os.geteuid() == 0:
                return True

            qualifiers.append(' (not required if run as root)')

        if self.if_not_systemd:
            if Path('/run/systemd/system').exists():
                return True

            qualifiers.append(' (not required if init is systemd)')

        if shutil.which(self.exe):
            return True

        if self.fatal:
            level = 'ERROR'
        else:
            level = 'WARNING'

        sys.stderr.write(
            '%s: command %s not found, provided by package %s%s\n' %
            (level, self.exe, self.package, ''.join(qualifiers))
        )
        return not self.fatal


class FileDependency(Dependency):
    def __init__(
        self,
        path: Path,
        package: str,
        *,
        fatal: bool = True,
    ) -> None:
        self.fatal = fatal
        self.path = path
        self.package = package

    def check(self) -> bool:
        if self.path.exists():
            return True

        if self.fatal:
            level = 'ERROR'
        else:
            level = 'WARNING'

        sys.stderr.write(
            '%s: file %s not found, provided by package %s\n' %
            (level, self.path, self.package)
        )
        return not self.fatal


class DirectoryDependency(Dependency):
    def __init__(
        self,
        path: Path
    ) -> None:
        self.path = path

    def check(self) -> bool:
        if not os.path.exists(self.path):
            sys.stderr.write(
                'INFO: directory %s not found, creating it now\n' %
                self.path
            )
            try:
                os.makedirs(self.path)
            except Exception as e:
                sys.stderr.write(
                    'ERROR: could not create directory: %s\n' % e
                )
                return False

        if not os.access(self.path, os.W_OK):
            sys.stderr.write(
                'ERROR: no write access to %s\n' % self.path
            )
            return False

        if not os.path.isdir(self.path):
            sys.stderr.write(
                'ERROR: %s is not a directory\n' % self.path
            )
            return False

        return True


class KvmDependency(Dependency):
    def __init__(
        self,
        *,
        if_exists: bool,
    ) -> None:
        self.if_exists = if_exists

    def check(self) -> bool:
        if os.access('/dev/kvm', os.W_OK):
            return True

        if not os.path.exists('/dev/kvm'):
            if self.if_exists:
                return True
            else:
                sys.stderr.write('ERROR: /dev/kvm does not exist\n')
                return False

        sys.stderr.write('ERROR: no permission to write /dev/kvm\n')
        return False


class SpaceRequirement(Dependency):
    '''Check if disk at path has specified amount of free space'''
    def __init__(
            self,
            path: Path,
            space: int,
            *,
            fatal: bool = False,
    ) -> None:
        self.path = path
        self.space = space
        self.fatal = fatal

    def check(self) -> bool:
        if not os.path.exists(self.path):
            return False
        free_space = shutil.disk_usage(self.path).free

        if self.fatal:
            level = 'ERROR'
        else:
            level = 'WARNING'

        if free_space <= self.space:
            missing = self.space - free_space
            sys.stderr.write('%s: Missing %s bytes of space in directory %s \n'
                             % (level, missing, os.path.realpath(self.path)))
            return not self.fatal
        return True


class Package(Dependency):
    def __init__(
            self,
            name: str,
            fatal: bool = False
    ):
        self.name = name
        self.fatal = fatal

    def check(self) -> bool:
        level = "ERROR" if self.fatal else "WARNING"
        dpkg_ret = subprocess.run(["dpkg-query", "--showformat=${Status}", "--show", self.name],
                                  stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                                  text=True)
        if dpkg_ret.stdout != "install ok installed":
            sys.stderr.write("%s: Missing package %s\n" % (level, self.name))
            return not self.fatal

        return True


def check_dependencies(dependencies: Sequence[Dependency]):
    ok = True

    # Intentionally not short-circuiting: we want to list all missing
    # dependencies at once
    for dep in dependencies:
        if not dep.check():
            ok = False

    return ok



#!/usr/bin/python3
#
# This is not a stable API; for use within autopkgtest only.
#
# Part of autopkgtest.
# autopkgtest is a tool for testing Debian binary packages.
#
# Copyright 2006-2016 Canonical Ltd.
# Copyright 2016-2020 Simon McVittie
# Copyright 2017 Martin Pitt
# Copyright 2017 Jiri Palecek
# Copyright 2017-2018 Collabora Ltd.
# Copyright 2018 Thadeu Lima de Souza Cascardo
# Copyright 2019 Michael Biebl
# Copyright 2019 Raphaël Hertzog
#
# autopkgtest-virt-qemu was developed by
# Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import errno
import fcntl
import json
import os
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import (
    List,
    Literal,
    Optional,
    Sequence,
    Union,
)

import VirtSubproc
import adtlog


class AllocatedPort(int):
    '''This object represents a port number. However this port number also
    happens to be locked using POSIX lockf. If a process owning an
    AllocatedPort dies in a bad way, the lock is released, but the file
    is left behind. If the AllocatedPort is release()ed the file is gone as
    well.
    '''
    _lockfd: Optional[int]
    _lockfile: Optional[str]

    def __new__(cls, value: int, lockfd: int, lockfile: str) -> "AllocatedPort":
        self = int.__new__(AllocatedPort, value)
        self._lockfd = lockfd
        self._lockfile = lockfile
        return self

    def release(self) -> None:
        if self._lockfile is not None:
            os.unlink(self._lockfile)
            self._lockfile = None
        if self._lockfd is not None:
            os.close(self._lockfd)
            self._lockfd = None

    __del__ = release


def find_free_port(start: int) -> Union[AllocatedPort, Literal[0]]:
    '''Find an unused port in the range [start, start+50)
    Returns an AllocatedPort. The port reservation is valid until that
    object has been .release()d or garbage collected.
    '''
    fd = None
    try:
        for p in range(start, start + 50):
            adtlog.debug('find_free_port: trying %i' % p)
            lockfile = '/tmp/autopkgtest-virt-qemu.port.%i' % p
            try:
                fd = os.open(lockfile, os.O_RDWR | os.O_CREAT)
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                # There is a little race condition. We may have opened
                # a previous file and managed to lock it, but
                # AllocatedPort.release may have gotten to delete it.
                if not os.access(lockfile, os.F_OK):
                    os.close(fd)
                    fd = None
                    adtlog.debug(f'find_free_port: lost race for {p}')
                    continue
                # The lock is now held until we close fd.
            except (IOError, OSError):
                if fd is not None:
                    os.close(fd)
                    fd = None
                adtlog.debug('find_free_port: %i is locked' % p)
                continue

            try:
                s = socket.create_connection(('127.0.0.1', p))
                # if that works, the port is taken
                s.close()
                continue
            except socket.error as e:
                if e.errno == errno.ECONNREFUSED:
                    adtlog.debug('find_free_port: %i is free' % p)
                    ret = AllocatedPort(p, fd, lockfile)
                    fd = None
                    return ret
                os.close(fd)
                fd = None
    finally:
        if fd is not None:
            os.close(fd)

    adtlog.debug('find_free_port: all ports are taken')
    return 0


def get_cpuflag() -> Sequence[str]:
    '''Return QEMU cpu option list suitable for host CPU'''

    try:
        with open('/proc/cpuinfo', 'r') as f:
            for line in f:
                if line.startswith('flags'):
                    words = line.split()
                    if 'vmx' in words:
                        adtlog.debug(
                            'Detected KVM capable Intel host CPU, '
                            'enabling nested KVM'
                        )
                        return ['-cpu', 'kvm64,+vmx,+lahf_lm']
                    elif 'svm' in words:  # AMD kvm
                        adtlog.debug(
                            'Detected KVM capable AMD host CPU, '
                            'enabling nested KVM'
                        )
                        # FIXME: this should really be the one below
                        # for more reproducible testbeds, but nothing
                        # except -cpu host works
                        # return ['-cpu', 'kvm64,+svm,+lahf_lm']
                        return ['-cpu', 'host']
    except IOError as e:
        adtlog.warning(
            'Cannot read /proc/cpuinfo to detect CPU flags: %s' % e
        )
        # fetching CPU flags isn't critical (only used to enable
        # nested KVM), so don't fail here

    return []


def get_host_time_zone() -> str:
    '''Return the host system's time zone, or an empty string.'''
    try:
        target = os.readlink('/etc/localtime')

        if '/zoneinfo/' in target:
            return target.split('/zoneinfo/')[-1]
    except OSError:
        pass

    try:
        with open('/etc/timezone', 'r') as f:
            for line in f:
                if line.startswith('#'):
                    continue
                line = line.strip()
                if line:
                    return line
    except OSError:
        pass

    return ''


class QemuImage:
    def __init__(
        self,
        file: str,
        format: Optional[str] = None,
        readonly: bool = False,
    ) -> None:
        self.file = file
        self.overlay: Optional[str] = None
        self.readonly = readonly

        if format is not None:
            self.format = format
        else:
            info = json.loads(
                VirtSubproc.check_exec([
                    'qemu-img', 'info',
                    '--output=json',
                    self.file,
                ], outp=True, timeout=5)
            )

            if 'format' not in info:
                VirtSubproc.bomb('Unable to determine format of %s' % self.file)

            self.format = str(info['format'])

    def __str__(self) -> str:
        bits: List[str] = []

        if self.overlay is None:
            bits.append('file={}'.format(self.file))
            bits.append('format={}'.format(self.format))
        else:
            bits.append('file={}'.format(self.overlay))
            bits.append('format=qcow2')
            bits.append('cache=unsafe')

        bits.append('if=virtio')
        bits.append('discard=unmap')

        if self.readonly:
            bits.append('readonly=on')

        return ','.join(bits)


class QemuSession:
    '''
    A running instance of qemu.
    '''

    def __init__(
        self,
        images: Sequence[Union[QemuImage, str]],
        *,
        boot: str,
        cpus: int = 1,
        efi_code: Optional[Path] = None,
        efi_data: Optional[Path] = None,
        overlay: bool = False,
        overlay_dir: Optional[str] = None,
        qemu_architecture: str,
        qemu_command: str,
        qemu_options: Sequence[str] = (),
        ram_size: int = 1024,
        workdir: Optional[str] = None
    ) -> None:
        """
        Constructor.

        images: Disk images for the VM. The first image is assumed to be
            the bootable, writable root filesystem, and we actually boot a
            snapshot. The remaining images are assumed to be read-only.

        boot: bios, efi, ieee1275 or none
        cpus: Number of vCPUs
        efi_code: Path to e.g. /usr/share/OVMF/OVMF_CODE_4M.fd
        efi_data: Path to e.g. /usr/share/OVMF/OVMF_VARS_4M.fd
        overlay: If true, use a temporary overlay for first image
        overlay_dir: Store writable overlays here (default: workdir)
        qemu_architecture: Architecture, from qemu's vocabulary
        qemu_command: qemu executable
        qemu_options: Space-separated options for qemu
        ram_size: Amount of RAM in MiB
        workdir: Directory for temporary files (default: a random
            subdirectory of $TMPDIR)
        """

        self.consoles = set(['hvc0', 'hvc1'])
        self.cpus = cpus
        self.images: List[QemuImage] = []
        self.overlay_dir = overlay_dir
        self.qemu_architecture = qemu_architecture
        self.qemu_command = qemu_command
        self.ram_size = ram_size
        self.ssh_port = find_free_port(10022)

        if workdir is None:
            workdir = tempfile.mkdtemp(prefix='autopkgtest-qemu.')

        self.workdir: Optional[str] = workdir
        os.chmod(workdir, 0o755)
        self.shareddir = os.path.join(workdir, 'shared')
        os.mkdir(self.shareddir)

        for i, image in enumerate(images):
            if isinstance(image, QemuImage):
                self.images.append(image)
            else:
                assert isinstance(image, str)

                self.images.append(
                    QemuImage(
                        file=image,
                        format=None,
                        readonly=(i != 0),
                    )
                )

        if overlay:
            self.images[0].overlay = self.prepare_overlay(self.images[0])

        if self.ssh_port:
            adtlog.debug(
                'Forwarding local port %i to VM ssh port 22' % self.ssh_port
            )
            nic_opt = ',hostfwd=tcp:127.0.0.1:%i-:22' % self.ssh_port
        else:
            nic_opt = ''

        # start QEMU
        argv = [
            self.qemu_command,
        ]

        if '-machine' not in qemu_options:
            # Some architectures do not have a default machine model
            if self.qemu_architecture == 'aarch64':
                # gic-version=max is necessary to enable kvm support when
                # running natively on an arm64 host
                if os.uname()[4] == 'aarch64':
                    argv.extend(['-machine', 'type=virt,gic-version=max'])
                else:
                    argv.extend(['-machine', 'virt'])
            elif self.qemu_architecture == 'arm':
                argv.extend(['-machine', 'virt'])
            elif self.qemu_architecture == 'riscv64':
                argv.extend(['-machine', 'virt,acpi=off'])
            elif (
                self.qemu_architecture in ('i386', 'x86_64') and
                boot == 'efi'
            ):
                argv.extend(['-machine', 'q35'])

        # Some architectures can only be run with certain CPUs
        if '-cpu' not in qemu_options:
            if self.qemu_architecture == 'aarch64':
                if os.uname()[4] == 'aarch64':
                    # Specifying a particular CPU doesn't seem to be
                    # possible when running natively with kvm
                    argv.extend(['-cpu', 'host'])
                else:
                    # -cpu host is obviously not possible on a foreign
                    # CPU, -cpu max is unusably slow, but this works!
                    argv.extend(['-cpu', 'cortex-a53'])

        argv.extend([
            '-m', str(ram_size),
            '-smp', str(cpus),
            '-nographic',
            '-device', 'virtio-net-pci,netdev=net0',
            '-netdev', 'user,id=net0' + nic_opt,
            '-object', 'rng-random,filename=/dev/urandom,id=rng0',
            '-device', 'virtio-rng-pci,rng=rng0,id=rng-device0',
            '-monitor',
            'unix:%s,server=on,wait=off' % self.get_socket_path('monitor'),
            '-virtfs',
            (
                'local,id=autopkgtest,path=%s,security_model=none,'
                'mount_tag=autopkgtest'
            ) % self.shareddir,
        ])

        for hvc in ('hvc0', 'hvc1'):
            if hvc == 'hvc0' and self.qemu_architecture == 'ppc64le':
                # The first -serial argument shows up as /dev/hvc0 in
                # the VM on ppc64le, so identify it as such
                argv.extend([
                    '-serial',
                    'unix:%s,server=on,wait=off' % self.get_socket_path('hvc0'),
                ])
            else:
                if hvc == 'hvc0' or self.qemu_architecture == 'ppc64le':
                    argv.extend([
                        '-device', 'virtio-serial',
                    ])
                argv.extend([
                    '-chardev',
                    'socket,path=%s,server=on,wait=off,id=%s' % (
                        self.get_socket_path(hvc), hvc,
                    ),
                    '-device', 'virtconsole,chardev=%s' % hvc,
                ])

        if self.qemu_architecture in ('x86_64', 'i386'):
            self.consoles.add('ttyS0')
            self.consoles.add('ttyS1')
            argv.extend([
                '-serial',
                'unix:%s,server=on,wait=off' % self.get_socket_path('ttyS0'),
                '-serial',
                'unix:%s,server=on,wait=off' % self.get_socket_path('ttyS1'),
            ])
        elif self.qemu_architecture == 'ppc64le':
            # There are no emulated hardware serial ports in this qemu
            # implementation, only hypervisor virtual consoles
            pass
        else:
            # On ARM this ends up as ttyAMA0 rather than ttyS0, but that's
            # close enough.
            self.consoles.add('ttyS0')
            argv.extend([
                '-serial',
                'unix:%s,server=on,wait=off' % self.get_socket_path('ttyS0'),
            ])

        for i, image in enumerate(self.images):
            argv.append('-drive')
            argv.append('index=%d,%s' % (i, image))

        if efi_code is not None:
            argv.append('-drive')
            argv.append(
                f'if=pflash,format=raw,unit=0,read-only=on,file={efi_code}'
            )

        if efi_data is not None:
            shutil.copy(str(efi_data), '%s/efivars.fd' % workdir)
            argv.append('-drive')
            argv.append(
                'if=pflash,format=raw,unit=1,file=%s/efivars.fd' % workdir
            )

        if os.path.exists('/dev/kvm'):
            if QemuFactory.kvm_compatible(self.qemu_architecture):
                argv.append('-enable-kvm')

            # Enable nested KVM by default on x86_64
            if (
                os.uname()[4] == 'x86_64' and
                self.qemu_architecture == 'x86_64' and
                '-cpu' not in qemu_options
            ):
                argv += get_cpuflag()

        # pass through option to qemu
        if qemu_options:
            argv.extend(qemu_options)

        adtlog.debug('full qemu command-line: %s' % argv)

        self.subprocess: Optional[subprocess.Popen[bytes]] = subprocess.Popen(
            argv,
            stdin=subprocess.DEVNULL,
            stdout=sys.stderr,
            stderr=subprocess.STDOUT,
        )

    def get_socket_path(self, name: str) -> str:
        assert self.workdir is not None
        return os.path.join(self.workdir, name)

    @property
    def monitor_socket(self) -> socket.socket:
        return VirtSubproc.get_unix_socket(self.get_socket_path('monitor'))

    def get_console_socket(self) -> socket.socket:
        for name in ('ttyS0', 'hvc0'):
            if name in self.consoles:
                path = self.get_socket_path(name)
                return VirtSubproc.get_unix_socket(path)

        VirtSubproc.bomb('No console socket available')
        raise AssertionError            # not reached

    def prepare_overlay(
        self,
        image: QemuImage,
    ) -> str:
        '''Generate a temporary overlay image'''

        # generate a temporary overlay
        if self.overlay_dir is not None:
            overlay = os.path.join(
                self.overlay_dir,
                os.path.basename(image.file) + '.overlay-%s' % time.time()
            )
        else:
            workdir = self.workdir
            assert workdir is not None
            overlay = os.path.join(workdir, 'overlay.img')

        adtlog.debug('Creating temporary overlay image in %s' % overlay)
        VirtSubproc.check_exec(
            [
                'qemu-img', 'create',
                '-f', 'qcow2',
                '-F', image.format,
                '-b', os.path.abspath(image.file),
                overlay,
            ],
            outp=True,
            timeout=300,
        )
        return overlay

    def cleanup(self, timeout_poweroff: Optional[int] = None) -> Optional[int]:
        ret = None

        if self.subprocess is not None:
            if timeout_poweroff is not None:
                try:
                    self.subprocess.wait(timeout_poweroff)
                except subprocess.TimeoutExpired:
                    pass
            self.subprocess.terminate()
            ret = self.subprocess.wait()
            self.subprocess = None

        if self.workdir is not None:
            shutil.rmtree(self.workdir)
            self.workdir = None

        return ret


class QemuFactory:
    '''
    A factory object that decides which qemu command is required,
    and can start a qemu session.
    '''

    def __init__(
        self,
        *,
        boot: str = 'auto',
        dpkg_architecture: Optional[str] = None,
        qemu_architecture: Optional[str] = None,
        qemu_command: Optional[str] = None,
        qemu_options: Sequence[str] = (),
    ) -> None:
        '''
        Constructor.

        boot: auto, bios, efi, ieee1275 or none
        dpkg_architecture: Architecture, from dpkg's vocabulary
        qemu_architecture: Architecture, from qemu's vocabulary
        qemu_command: qemu executable
        qemu_options: Arbitrary options for qemu
        '''

        if qemu_architecture is not None:
            self.qemu_architecture: str = qemu_architecture
        elif dpkg_architecture is not None:
            self.qemu_architecture = self.qemu_arch_for_dpkg_arch(
                dpkg_architecture
            )
        else:
            self.qemu_architecture = ''

        if not self.qemu_architecture and shutil.which('dpkg') is not None:
            dpkg_architecture = subprocess.check_output(
                ['dpkg', '--print-architecture'],
                universal_newlines=True
            ).strip()
            self.qemu_architecture = self.qemu_arch_for_dpkg_arch(
                dpkg_architecture
            )

        if not self.qemu_architecture and qemu_command is not None:
            guess = self.qemu_arch_for_command(qemu_command)

            if guess is not None:
                self.qemu_architecture = guess

        if not self.qemu_architecture:
            self.qemu_architecture = self.qemu_arch_for_uname()

        self.arm_on_aarch64 = False

        if qemu_command is not None:
            self.qemu_command: str = qemu_command
        elif (
            self.qemu_architecture == 'arm' and
            os.uname()[4] == 'aarch64'
        ):
            # aarch64 host systems can't kvm-accelerate qemu-system-arm,
            # so automatically switch to qemu-system-aarch64 in 32-bit mode
            # (emulating a 32-bit ARMv8a)
            self.qemu_architecture = 'aarch64'
            self.qemu_command = 'qemu-system-aarch64'
            self.arm_on_aarch64 = True
        else:
            self.qemu_command = self.qemu_command_for_arch(
                self.qemu_architecture
            )

        if boot == 'auto':
            if self.qemu_architecture in ('i386', 'x86_64'):
                # virt-qemu doesn't do anything special for BIOS boot,
                # but build-qemu does
                boot = 'bios'
            elif self.qemu_architecture in ('arm', 'aarch64', 'riscv64'):
                # Handled specially below
                boot = 'efi'
            elif self.qemu_architecture == 'ppc64le':
                # virt-qemu doesn't do anything special for IEEE 1275
                # (OpenFirmware) boot, but build-qemu does
                boot = 'ieee1275'
            else:
                adtlog.debug(
                    'autopkgtest does not know how to boot this machine, '
                    'hopefully nothing special needs to be done?'
                )
                boot = 'none'

        self.boot = boot
        self.efi_code: Optional[Path] = None
        self.efi_data: Optional[Path] = None
        self.efi_package: str = ''

        if boot == 'efi':
            if self.qemu_architecture == 'x86_64':
                code = '/usr/share/OVMF/OVMF_CODE_4M.fd'
                data = '/usr/share/OVMF/OVMF_VARS_4M.fd'
                package = 'ovmf'

                if not os.path.exists(code):
                    code = '/usr/share/OVMF/OVMF_CODE.fd'
                    data = '/usr/share/OVMF/OVMF_VARS.fd'
            elif self.qemu_architecture == 'i386':
                code = '/usr/share/OVMF/OVMF32_CODE_4M.secboot.fd'
                data = '/usr/share/OVMF/OVMF32_VARS_4M.fd'
                package = 'ovmf-ia32'
            elif (
                self.qemu_architecture == 'arm' or
                self.arm_on_aarch64 or (
                    self.qemu_architecture == 'aarch64' and
                    'aarch64=off' in str(qemu_options)
                )
            ):
                code = '/usr/share/AAVMF/AAVMF32_CODE.fd'
                data = '/usr/share/AAVMF/AAVMF32_VARS.fd'
                package = 'qemu-efi-arm'
            elif self.qemu_architecture == 'aarch64':
                code = '/usr/share/AAVMF/AAVMF_CODE.fd'
                data = '/usr/share/AAVMF/AAVMF_VARS.fd'
                package = 'qemu-efi-aarch64'
            elif self.qemu_architecture == 'riscv64':
                code = '/usr/share/qemu-efi-riscv64/RISCV_VIRT_CODE.fd'
                data = '/usr/share/qemu-efi-riscv64/RISCV_VIRT_VARS.fd'
                package = 'qemu-efi-riscv64'
            else:
                VirtSubproc.bomb('Unknown architecture for EFI boot')

            self.efi_code = Path(code)
            self.efi_data = Path(data)
            self.efi_package = package
        else:
            adtlog.debug(
                'Assuming nothing special needs to be done to set up '
                'firmware to boot this machine (boot method: %s)' % boot
            )

        self.qemu_options = list(qemu_options)

    def new_session(
        self,
        **kwargs
    ) -> QemuSession:
        adtlog.debug('qemu architecture: %s' % self.qemu_architecture)
        adtlog.debug('qemu command: %s' % self.qemu_command)

        qemu_options = self.qemu_options

        if self.arm_on_aarch64 and '-cpu' not in qemu_options:
            qemu_options = [
                '-cpu', 'host,aarch64=off'
            ] + qemu_options

        return QemuSession(
            boot=self.boot,
            efi_code=self.efi_code,
            efi_data=self.efi_data,
            qemu_architecture=self.qemu_architecture,
            qemu_command=self.qemu_command,
            qemu_options=qemu_options,
            **kwargs,
        )

    @staticmethod
    def dpkg_arch_for_qemu_arch(qemu_architecture: str) -> str:
        return {
            'aarch64': 'arm64',
            # alpha
            'arm': 'armhf',     # or armel or arm
            'avr': 'avr32',
            # cris not in dpkg-architecture -L
            # hppa
            # i386
            'loongarch64': 'loong64',
            # m68k
            # microblaze not in dpkg-architecture -L
            # microblazeel not in dpkg-architecture -L
            # mips
            # mips64
            # mips64el
            # mipsel
            # moxie not in dpkg-architecture -L
            # nios2
            # or1k
            'ppc': 'powerpc',
            # ppc64
            'ppc64le': 'ppc64el',
            # riscv32 not in dpkg-architecture -L
            # riscv64
            # rx not in dpkg-architecture -L
            # s390x
            # sh4
            # sh4eb
            # sparc
            # sparc64
            # tricore not in dpkg-architecture -L
            'x86_64': 'amd64',          # or x32
            'x86_64-microvm': 'amd64',  # or x32
            # xtensa not in dpkg-architecture -L
            # xtensaeb not in dpkg-architecture -L
        }.get(qemu_architecture, qemu_architecture)

    @staticmethod
    def qemu_arch_for_command(qemu_command: str) -> Optional[str]:
        known = [
            'aarch64',
            'alpha',
            'arm',
            'avr',
            'cris',
            'hppa',
            'i386',
            'loongarch64',
            'm68k',
            'microblaze',
            'microblazeel',
            'mips',
            'mips64',
            'mips64el',
            'mipsel',
            'moxie',
            'nios2',
            'or1k',
            'ppc',
            'ppc64',
            'ppc64le',
            'riscv32',
            'riscv64',
            'rx',
            's390x',
            'sh4',
            'sh4eb',
            'sparc',
            'sparc64',
            'tricore',
            'x86_64',
            'x86_64-microvm',
            'xtensa',
            'xtensaeb',
        ]

        base = os.path.basename(qemu_command)

        for k in known:
            if base == 'qemu-system-' + k:
                return k

        for k in known:
            if ('qemu-system-' + k) in base:
                return k

        return None

    # TODO: when we require qemu 8 (Debian 13, Ubuntu 24.04) we will be
    # able to rely on each dpkg architecture having a symlink
    # qemu-system-${dpkg_architecture} -> qemu-system-${qemu_architecture}
    # if different, allowing this to be simplified.
    @staticmethod
    def qemu_arch_for_dpkg_arch(dpkg_architecture: str) -> str:
        return {
            # alpha
            'amd64': 'x86_64',
            # arm
            'arm64': 'aarch64',
            'arm64ilp32': 'aarch64',
            # armeb?
            'armel': 'arm',
            'armhf': 'arm',
            'avr32': 'avr',
            # hppa
            # i386
            # ia64?
            'loong64': 'loongarch64',
            # m32r?
            # m68k
            # mips
            # mips64
            # mips64el
            # mipsel
            # mipsn32?
            # mipsn32el?
            # mipsn32r6?
            # mipsn32r6el?
            # mipsr6?
            # mipsr6el?
            # nios2
            # or1k
            'powerpc': 'ppc',
            'powerpcel': 'ppc64le',     # ?
            # powerpcspe?
            # ppc64
            'ppc64el': 'ppc64le',
            # riscv64
            's390': 's390x',            # ?
            # s390x
            # sh3?
            # sh3eb?
            # sh4
            # sh4eb
            # sparc
            # sparc64
            # tilegx?
            'x32': 'x86_64',
        }.get(dpkg_architecture, dpkg_architecture)

    @staticmethod
    def qemu_arch_for_uname(uname_m: Optional[str] = None) -> str:
        uname_to_qemu_arch = {'i[3456]86$': 'i386', '^arm': 'arm'}

        if uname_m is None:
            uname_m = os.uname()[4]

        for pattern, arch in uname_to_qemu_arch.items():
            if re.match(pattern, uname_m):
                return arch
        else:
            # usually the qemu architecture is the same as Linux uname -m
            return uname_m

    @staticmethod
    def qemu_command_for_arch(qemu_architecture: str) -> str:
        return 'qemu-system-' + qemu_architecture

    @staticmethod
    def kvm_compatible(
        qemu_architecture: str,
        uname_m: Optional[str] = None
    ) -> bool:
        if uname_m is None:
            uname_m = os.uname()[4]

        if re.match('^i[3456]86$', uname_m):
            uname_m = 'i386'

        return qemu_architecture in {
            'x86_64': ['x86_64', 'i386'],
            'i386': ['i386'],
            'aarch64': ['aarch64'],
            # According to https://wiki.alpinelinux.org/wiki/S390x/Installation
            's390x': ['s390x'],
            'ppc64le': ['ppc64le'],
        }.get(uname_m, [])



'''Provide a fake package archive for testing'''

# (C) 2012-2015 Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import tempfile
import shutil
import os
import subprocess
import atexit
import textwrap


class Archive:
    def __init__(
            self,
            *,
            path=None,
            pooldir='pool',
            suite=None,
            codename=None,
            component=None,
            notautomatic=False,
            butautomaticupgrades=False,
            codename_symlink=False,
    ):
        '''Construct a local package test archive.

        By default, Packages.gz is created in the root of the target directory,
        so that the apt source will just use '/' for the suite and no
        components. If suite, codename and component are given, Packages.gz
        will be placed into 'dists/suite/component/binary-<arch>/', and the
        Suite and Codename fields of the Release file of this archive will be
        set to the specified values.

        If the notautomatic argument is true, "NotAutomatic: yes" will be set
        in the Release file. In a similar fashion, the butautomaticupgrades
        argument controls ButAutomaticUpgrades.

        If codename_symlink is true, the archive will be accessible by using
        the specified codename in the APT sources. This is accomplished by
        creating a dists/codename -> dists/suite symlink. This is useful to
        create archives with a Debian-like layout, where "unstable" is also
        available via its codename "sid".

        The archive is initially empty. You can create new packages with
        create_deb(). self.path contains the path of the archive, and
        self.apt_source provides an apt source "deb" line.

        If path is None (default), it is kept in a temporary directory which
        gets removed when the Archive object gets deleted. Otherwise the given
        path is used, which is useful for creating multiple suites/components;
        then you should use a different pooldir.
        '''
        if path:
            self.path = path
        else:
            self.path = tempfile.mkdtemp(prefix='testarchive.')
            atexit.register(shutil.rmtree, self.path)
        self.pooldir = pooldir

        arch = subprocess.check_output(['dpkg', '--print-architecture'],
                                       universal_newlines=True).strip()
        self.suite = suite
        self.codename = codename
        self.component = component
        self.notautomatic = notautomatic
        self.butautomaticupgrades = butautomaticupgrades
        if suite or codename or component:
            assert (
                suite and codename and component
            ), 'must specify all of none of: suite, codename, component'
            self.index_dir = os.path.join(
                'dists', suite, component, 'binary-' + arch)
            os.makedirs(os.path.join(self.path, self.index_dir))
            self.apt_source = 'deb [trusted=yes arch=%s] file://%s %s %s' % \
                (arch, self.path, suite, component)

            if codename_symlink and codename != suite:
                os.symlink(suite, os.path.join(self.path, "dists", codename))
        else:
            self.apt_source = 'deb [trusted=yes arch=%s] file://%s /' % \
                (arch, self.path)
            self.index_dir = ''

    def create_deb(self, name, version='1', architecture='all',
                   dependencies={}, description='test package', extra_tags={},
                   files={}, component='main', srcpkg=None, update_index=True):
        '''Build a deb package and add it to the archive.

        The only mandatory argument is the package name. You can additionally
        specify the package version (default '1'), architecture (default
        'all'), a dictionary with dependencies (empty by default; for example
        {'Depends': 'foo, bar', 'Conflicts: baz'}, a short description
        (default: 'test package'), and arbitrary extra tags.

        By default the package is empty. It can get files by specifying a
        path -> contents dictionary in 'files'. Paths must be relative.
        Example: files={'etc/foo.conf': 'enable=true'}

        The newly created deb automatically gets added to the "Packages" index,
        unless update_index is False.

        Return the path to the newly created deb package, in case you only need
        the deb itself, not the archive.
        '''
        d = tempfile.mkdtemp()
        os.mkdir(os.path.join(d, 'DEBIAN'))
        with open(os.path.join(d, 'DEBIAN', 'control'), 'w') as f:
            f.write('''Package: %s
Maintainer: Test User <test@example.com>
Version: %s
Priority: optional
Section: devel
Architecture: %s
''' % (name, version, architecture))

            if srcpkg:
                f.write('Source: %s\n' % srcpkg)

            for k, v in dependencies.items():
                f.write('%s: %s\n' % (k, v))

            f.write('''Description: %s
 Test dummy package.
''' % description)

            for k, v in extra_tags.items():
                f.write('%s: %s\n' % (k, v))

        for path, contents in files.items():
            if isinstance(contents, bytes):
                mode = 'wb'
            else:
                mode = 'w'
            pathdir = os.path.join(d, os.path.dirname(path))
            if not os.path.isdir(pathdir):
                os.makedirs(pathdir)
            with open(os.path.join(d, path), mode) as f:
                f.write(contents)

        if srcpkg is None:
            srcpkg = name
        if srcpkg.startswith('lib'):
            prefix = srcpkg[:4]
        else:
            prefix = srcpkg[0]
        dir = os.path.join(self.path, self.pooldir, component, prefix, srcpkg)
        if not os.path.isdir(dir):
            os.makedirs(dir)

        debpath = os.path.join(dir, '%s_%s_%s.deb' % (name, version,
                                                      architecture))
        subprocess.check_call(['dpkg-deb', '-Zgzip', '-b', d, debpath],
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        shutil.rmtree(d)
        assert os.path.exists(debpath)

        if update_index:
            self.update_index()

        return debpath

    def add_sources(self, name, binaries, version='1', component='main'):
        '''Add source package entry to the Sources index'''

        if not type(binaries) is dict:
            binaries = {k: 'any' for k in binaries}

        if name.startswith('lib'):
            prefix = name[:4]
        else:
            prefix = name[0]
        poolsubdir = os.path.join(self.pooldir, self.component, prefix, name)

        # create a dsc
        dir = os.path.join(self.path, self.pooldir, component, prefix, name)
        if not os.path.isdir(dir):
            os.makedirs(dir)
        dscpath = os.path.join(dir, '%s_%s.dsc' % (name, version))
        with open(dscpath, mode='w'):
            # For now we only care about the file being present,
            # so let's create it empty.
            pass

        if self.index_dir:
            srcindex_dir = os.path.join(self.path,
                                        os.path.dirname(self.index_dir),
                                        'source')
        else:
            srcindex_dir = self.index_dir
        if not os.path.isdir(srcindex_dir):
            os.makedirs(srcindex_dir)

        with open(os.path.join(srcindex_dir, 'Sources'), 'a') as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    Package: {name}
                    Binary: {', '.join(binaries.keys())}
                    Version: {version}
                    Priority: optional
                    Architecture: any all
                    Format: 1.0
                    Directory: {poolsubdir}
                    Files:
                     d41d8cd98f00b204e9800998ecf8427e 0 {name}_{version}.dsc
                    Package-List:
                    """
                )
            )
            for b, a in binaries.items():
                f.write(' %s deb admin optional arch=%s\n' % (b, a))
            f.write('Standards-Version: 1.0\n\n')

    def update_index(self):
        '''Update the Packages index and Release file.

        This usually gets done automatically by create_deb(), but needs to be
        done if you manually copy debs into the archive or call create_deb with
        update_index==False.
        '''
        old_cwd = os.getcwd()
        try:
            os.chdir(self.path)
            with open(os.path.join(self.index_dir, 'Packages'), 'wb') as f:
                subprocess.check_call(['apt-ftparchive', 'packages',
                                       self.pooldir], stdout=f)
            if self.suite:
                rp = os.path.join(self.path, 'dists', self.suite, 'Release')
                try:
                    os.unlink(rp)
                except OSError:
                    pass
                release_cmd = ['apt-ftparchive', 'release',
                               '-o', 'APT::FTPArchive::Release::Suite=' + self.suite,
                               '-o', 'APT::FTPArchive::Release::Codename=' + self.codename]
                if self.notautomatic:
                    release_cmd += ['-o', 'APT::FTPArchive::Release::NotAutomatic=yes']
                if self.butautomaticupgrades:
                    release_cmd += ['-o', 'APT::FTPArchive::Release::ButAutomaticUpgrades=yes']
                release_cmd += [os.path.dirname(rp)]
                release = subprocess.check_output(release_cmd)
                with open(rp, 'wb') as f:
                    f.write(release)
        finally:
            os.chdir(old_cwd)
        subprocess.check_call(['chmod', '-R', 'a+rX', '--', self.path])


if __name__ == '__main__':
    r = Archive(suite='testy', component='main')
    r.create_deb('vanilla')
    r.create_deb('libvanilla0', srcpkg='vanilla')
    r.create_deb('chocolate', dependencies={'Depends': 'vanilla'})
    print(r.apt_source)
    r.add_sources('vanilla', ['vanilla', 'libvanilla0'])

    p = Archive(path=r.path, pooldir='pool-proposed', suite='testy-proposed',
                codename='testy', component='main')
    p.create_deb('vanilla', '2')
    r.create_deb('libvanilla0', '2', srcpkg='vanilla')
    p.create_deb('chocolate', '2', dependencies={'Depends': 'vanilla (>= 2)'})
    print(p.apt_source)
    p.add_sources('vanilla', ['vanilla', 'libvanilla0'], '2')

    subprocess.call(['bash', '-i'], cwd=r.path)



# adtlog is part of autopkgtest
# autopkgtest is a tool for testing Debian binary packages
#
# autopkgtest is Copyright (C) 2006-2014 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import sys
import time
import errno
import os

summary_stream = None
verbosity = 1  # 0: quiet (warning/error only), 1: info, 2: debug
enable_colors = None


def log(message, level, prefix='', timestamp=False, color=None):
    '''Write a log message to stderr'''

    if level > verbosity:
        return

    # needs lazy initialization as it may be redirected to tee
    global enable_colors
    if enable_colors is None:
        enable_colors = os.isatty(sys.stderr.fileno())

    head = sys.argv[0].split('/')[-1]
    if timestamp:
        head += ' [%s]: ' % time.strftime('%H:%M:%S')
    else:
        head += ': '

    if prefix:
        head += prefix + ': '

    out = (head + message + '\n').encode('UTF-8')

    if color is not None and enable_colors:
        out = b'\033[3' + chr(47 + color).encode() + b'm' + out + b'\033[0m'

    # we sometimes hit EAGAIN here, try a few times
    retries = 10
    while retries >= 0:
        try:
            sys.stderr.buffer.write(out)
            break
        except IOError as e:
            if e.errno == errno.EAGAIN:
                retries -= 1
                time.sleep(0.05)
            else:
                raise
    sys.stderr.buffer.flush()


def error(message):
    log(message, 0, prefix='ERROR', timestamp=True, color=2)


def warning(message):
    log(message, 0, prefix='WARNING', color=5)


def info(message):
    log(message, 1, timestamp=True, color=4)


def debug(message):
    log(message, 2, prefix='DBG', timestamp=False, color=8)


def psummary(m):
    if summary_stream is not None:
        summary_stream.write(m.encode('UTF-8'))
        summary_stream.write(b'\n')


def preport(m):
    sys.stdout.buffer.write(m.encode('UTF-8'))
    sys.stdout.buffer.write(b'\n')
    sys.stdout.buffer.flush()
    psummary(m)


def report(tname, result):
    preport('%-20s %s' % (tname, result))


#
# Errors
#

class TestbedFailure(RuntimeError):
    pass


class BadPackageError(RuntimeError):
    pass


class AutopkgtestError(RuntimeError):
    pass


def bomb(m):
    raise AutopkgtestError(m)


def badpkg(m):
    raise BadPackageError(m)


class AptError(Exception):
    pass


class AptPermanentError(AptError):
    pass


class AptDownloadError(AptError):
    pass



# VirtSubproc is part of autopkgtest
# autopkgtest is a tool for testing Debian binary packages
#
# autopkgtest is Copyright (C) 2006-2007 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# See the file CREDITS for a full list of credits information (often
# installed as /usr/share/doc/autopkgtest/CREDITS).

import __main__

from functools import partial
import sys
import os
from urllib.parse import quote as url_quote
from urllib.parse import unquote as url_unquote
import signal
import subprocess
import traceback
import errno
import fcntl
import time
import socket
import shutil
import shlex
from typing import (List, Optional, Union, Tuple)

import adtlog

# When running installed, this is /usr/share/autopkgtest.
# When running uninstalled, this is the source tree.
# Either way, it has a lib subdirectory from which virtualization
# backends can read extra files.
PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

progname = "<VirtSubproc>"
devnull_read = open('/dev/null', 'rb')
caller = __main__
copy_timeout = int(os.getenv('AUTOPKGTEST_VIRT_COPY_TIMEOUT', '300'))

downtmp_open = None  # downtmp after opening testbed
downtmp = None  # current downtmp (None after close)
auxverb = None  # prefix to run command argv in testbed
cleaning = False
in_mainloop = False


class Quit(RuntimeError):

    def __init__(self, ec, m):
        self.ec = ec
        self.m = m


class Timeout(RuntimeError):

    def __init__(self, timeout_secs, abort=False):
        self.timeout_secs = timeout_secs
        self.abort = abort


def alarm_handler(to, *a):
    raise Timeout(to)


def timeout_start(to):
    signal.signal(signal.SIGALRM, partial(alarm_handler, to))
    signal.alarm(to)


def timeout_stop():
    signal.alarm(0)


class FailedCmd(RuntimeError):

    def __init__(self, e):
        self.e = e


def bomb(m):
    if in_mainloop:
        raise Quit(12, progname + ": failure: %s" % m)
    else:
        sys.stderr.write(m)
        sys.stderr.write('\n')
        sys.exit(1)


def ok():
    print('ok')


def cmdnumargs(c, ce, nargs=0, noptargs=0):
    if len(c) < 1 + nargs:
        bomb("too few arguments to command `%s'" % ce[0])
    if noptargs is not None and len(c) > 1 + nargs + noptargs:
        bomb("too many arguments to command `%s'" % ce[0])


def cmd_capabilities(c, ce):
    cmdnumargs(c, ce)
    return caller.hook_capabilities()


def cmd_quit(c, ce):
    cmdnumargs(c, ce)
    raise Quit(0, '')


def cmd_close(c, ce):
    cmdnumargs(c, ce)
    if not downtmp:
        bomb("`close' when not open")
    cleanup()


def cmd_print_execute_command(c, ce):
    global auxverb

    cmdnumargs(c, ce)
    if not downtmp:
        bomb("`print-execute-command' when not open")
    return [','.join(map(url_quote, auxverb))]


def execute_timeout(instr, timeout, *popenargs, **popenargsk):
    '''Popen wrapper with timeout supervision

    If instr is given, it is fed into stdin, otherwise stdin will be /dev/null.

    Return (status, stdout, stderr)
    '''
    adtlog.debug('execute-timeout: ' + ' '.join(popenargs[0]))
    if instr is None:
        popenargsk['stdin'] = devnull_read
    else:
        instr = instr.encode('UTF-8')
    sp = subprocess.Popen(*popenargs,
                          **popenargsk)
    timeout_start(timeout)
    try:
        (out, err) = sp.communicate(instr)
        if out is not None:
            out = out.decode('UTF-8', 'replace')
        if err is not None:
            err = err.decode('UTF-8', 'replace')
    except Timeout:
        try:
            sp.kill()
            sp.wait()
        except OSError as e:
            adtlog.error('WARNING: Cannot kill timed out process %s: %s' %
                         (popenargs[0], e))
        raise
    timeout_stop()
    status = sp.wait()
    return (status, out, err)


def check_exec(argv, downp=False, outp=False, timeout=0, fail_on_stderr=True):
    '''Run successful command (argv list)

    Command must succeed (exit code 0) and not produce any stderr. If downp is
    True, command is run in testbed. If outp is True, stdout will be captured
    and returned. stdin is set to /dev/null.

    Returns stdout (or None if outp is False).
    '''
    global auxverb

    if downp:
        real_argv = auxverb + argv
    else:
        real_argv = argv
    if outp:
        stdout = subprocess.PIPE
    else:
        stdout = None

    (status, out, err) = execute_timeout(None, timeout, real_argv,
                                         stdout=stdout, stderr=subprocess.PIPE)

    if status:
        bomb("%s%s failed (exit status %d, stderr %r)" %
             ((downp and "(down) " or ""), argv, status, err))
    if fail_on_stderr and err:
        bomb("%s unexpectedly produced stderr output `%s'" %
             (argv, err))

    if outp and out and out[-1] == '\n':
        out = out[:-1]
    return out


def load_shell_script(path: str) -> str:
    if not os.path.isabs(path):
        path = os.path.join(PKGDATADIR, path)

    with open(path) as reader:
        return reader.read()


def wait_booted(
    exec_command: List[str],
    timeout: int = 120,
    short_command_timeout: int = 10,
) -> None:
    '''
    Wait until a container or VM with an init system has sufficiently
    booted to interact with it.

    timeout: how long to wait for the testbed to complete boot

    short_command_timeout: how long we will wait for a relatively quick command
    to finish
    '''
    await_boot = load_shell_script('lib/in-testbed/await-boot.sh')

    while timeout > 0:
        timeout -= 1
        time.sleep(1)
        adtlog.debug('wait_booted: trying to execute trivial command on container')
        try:
            (rc, _, _) = execute_timeout(
                None, short_command_timeout, exec_command + ['true'],
            )
        except Timeout:
            # With some virt servers, commands can get stuck when executed
            # while the testbed is booting, or at least this is what we observe.
            # The wait_booted() retry loop has its own timeout, so we shouldn't
            # stop on timeouts internal to the loop.
            rc = 1
        if rc == 0:
            adtlog.debug("wait_booted: trivial command executed successfully")
            break
    else:
        bomb('timed out waiting for testbed to start')

    adtlog.debug('testbed is alive, waiting for boot to complete')
    try:
        check_exec(
            exec_command + ['sh', '-ec', await_boot],
            timeout=timeout,
            fail_on_stderr=False,
        )
    except Timeout:
        bomb('timed out waiting for testbed to complete boot')


class timeout:
    def __init__(self, secs, exit_msg=None):
        '''Context manager that times out after given number of seconds.

        If exit_msg is given, the program bomb()s with that message,
        otherwise it raises a Timeout exception.
        '''
        self.secs = secs
        self.exit_msg = exit_msg

    def __enter__(self):
        timeout_start(self.secs)

    def __exit__(self, type_, value, traceback):
        timeout_stop()
        if type_ is Timeout and self.exit_msg:
            bomb(self.exit_msg)
            return True
        return False


def get_unix_socket(path):
    '''Open a connected client socket to given Unix socket with a 5s timeout'''

    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    with timeout(5, 'Timed out waiting for %s socket\n' % path):
        while True:
            try:
                s.connect(path)
                break
            except socket.error:
                continue
    return s


def matches_expectation(
    haystack: bytes,
    needle: Union[None, bytes, str, Tuple[Union[bytes, str]]],
) -> bool:
    if needle is None:
        return True
    elif isinstance(needle, tuple):
        for n in needle:
            if matches_expectation(haystack, n):
                return True
        else:
            return False
    elif isinstance(needle, bytes):
        return needle in haystack
    elif isinstance(needle, str):
        return needle.encode('utf-8') in haystack
    else:
        raise TypeError(repr(needle))


def expect(
    sock: socket.socket,
    search: Union[None, bytes, str, Tuple[Union[bytes, str]]],
    timeout_sec: int,
    description: Optional[str] = None,
    echo: bool = False,
):
    adtlog.debug('expect: %r' % (search,))
    what = repr(description or search or 'data')
    out = b''
    with timeout(timeout_sec,
                 description and ('timed out waiting for %s' % what) or None):
        while True:
            block = sock.recv(4096)
            if not block:
                time.sleep(0.1)
                continue
            if echo:
                sys.stderr.buffer.write(block)
            out += block

            if matches_expectation(out, search):
                adtlog.debug('expect: found "%s"' % what)
                break

    return out


def cmd_open(c, ce):
    global auxverb, downtmp, downtmp_open
    cmdnumargs(c, ce)
    if downtmp:
        bomb("`open' when already open")
    caller.hook_open()
    adtlog.debug("auxverb = %s, downtmp = %s" % (str(auxverb), downtmp))
    downtmp = caller.hook_downtmp(downtmp_open)
    if downtmp_open and downtmp_open != downtmp:
        bomb('virt-runner failed to restore downtmp path %s, gave %s instead'
             % (downtmp_open, downtmp))
    downtmp_open = downtmp
    return [downtmp]


def downtmp_mktemp(capabilities, path, downtmp_prefix):
    '''Generate a downtmp

    When a path is given, this is the downtmp that we created when opening the
    testbed the first time. We always want to keep the same path between
    resets, as built package trees sometimes refer to absolute paths and thus
    fail if they get moved around.

    capabilities will be used to determine the appropriate permissions.

    If downtmp_prefix is not None, updates capabilities to add an downtmp-host,
    consisting of downtmp_prefix + the downtmp path.
    '''

    # The caller is entitled to use the scratch directory normally, without taking
    # the kind of extreme special care needed for a world-writeable directory.
    if any(cap in ['isolation-container', 'isolation-machine'] for cap in capabilities):
        # The testbed is fully isolated by the host, so the only things running
        # in the testbed are part of the testbed or specified by the caller.
        # So there is no untrusted code.
        #
        # Make the scratch directory world-writeable is convenient for use cases
        # that have to deal with multiple users.
        #
        # If the downtmp is visible to the host, then the isolation is not enough.
        # This assertion is a double-check, therefore, that we aren't exposing
        # the caller's code (which is using the downtmp) to possible hostile actions.
        assert downtmp_prefix is None
        mode = '777'
    else:
        # Untrusted code might be running in the testbed - so we must protect it.
        mode = '755'

    if path:
        check_exec(['/bin/mkdir', f"--mode={mode}", '--parents', downtmp_open],
                   downp=True)
        d = path
    else:
        d = check_exec(['/bin/mktemp', '--directory', '--tmpdir', 'autopkgtest.XXXXXX'],
                       downp=True, outp=True)
        check_exec(['/bin/chmod', mode, d], downp=True)
    if downtmp_prefix is not None:
        capabilities.append('downtmp-host=' + downtmp_prefix + d)
    return d


def downtmp_remove(capabilities):
    '''Removes the downtmp

    Also removes any downtmp-host entry from capabilities.
    '''
    global downtmp, auxverb

    # Remove from capabilities first.  Some call sites catch exceptions and continue,
    # and they also want to see downtmp removed.
    capabilities[:] = [c for c in capabilities if not c.startswith('downtmp-host=')]

    if downtmp:
        execute_timeout(None, copy_timeout,
                        auxverb + ['rm', '-rf', '--', downtmp])
        downtmp = None


def cmd_revert(c, ce):
    global auxverb, downtmp, downtmp_open
    cmdnumargs(c, ce)
    if not downtmp:
        bomb("`revert' when not open")
    if 'revert' not in caller.hook_capabilities():
        bomb("`revert' when `revert' not advertised")
    caller.hook_revert()
    downtmp = caller.hook_downtmp(downtmp_open)
    if downtmp_open and downtmp_open != downtmp:
        bomb('virt-runner failed to restore downtmp path %s, gave %s instead'
             % (downtmp_open, downtmp))
    adtlog.debug("auxverb = %s, downtmp = %s" % (str(auxverb), downtmp))

    return [downtmp]


def reboot_testbed():
    (systemd_check, out, err) = execute_timeout(
        None, 10, auxverb + ['test', '-d', '/run/systemd/system'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    if systemd_check == 0:
        reboot_cmd = ['systemd-run', '--no-block', '--quiet', 'sh', '-c', 'sleep 3; reboot']
        adtlog.debug('rebooting with command: %s' % reboot_cmd)
        status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)

        if status == 0:
            return
        else:
            adtlog.debug(
                'Cannot reboot with %s, trying something else' % reboot_cmd)

    reboot_cmd = ['sh', '-c', '(sleep 3; reboot) >/dev/null 2>&1 &']
    adtlog.debug('rebooting with command: %s' % reboot_cmd)
    status, out, err = execute_timeout(None, 30, auxverb + reboot_cmd)


def cmd_reboot(c, ce):
    global downtmp
    cmdnumargs(c, ce, 0, 1)
    wait_reboot_args = {}
    if not downtmp:
        bomb("`reboot' when not open")
    if 'reboot' not in caller.hook_capabilities():
        bomb("`reboot' when `reboot' not advertised")

    # save current downtmp; try a few locations, as /var/cache might be r/o
    # (argh Ubuntu touch)
    directories = '/var/cache /home'
    check_exec(['/bin/sh', '-ec', 'for d in %s; do if [ -w $d ]; then '
                '  /bin/tar --warning=none --create --absolute-names '
                '''    -f $d/autopkgtest-tmpdir.tar '%s'; '''
                '  /bin/rm -f /run/autopkgtest-reboot-prepare-mark; '
                '  exit 0; fi; done; exit 1' % (directories, downtmp)],
               downp=True, timeout=copy_timeout)
    adtlog.debug('cmd_reboot: saved current downtmp, rebooting')

    try:
        wait_reboot_args = caller.hook_prepare_reboot() or {}
    except AttributeError:
        pass

    # reboot
    if len(c) > 1 and c[1] == 'prepare-only':
        adtlog.info('state saved, waiting for testbed to reboot...')
    else:
        reboot_testbed()

    caller.hook_wait_reboot(**wait_reboot_args)

    # restore downtmp
    check_exec(['/bin/sh', '-ec', 'for d in %s; do '
                'if [ -e $d/autopkgtest-tmpdir.tar ]; then '
                ' /bin/tar --warning=none --extract --absolute-names '
                '     -f $d/autopkgtest-tmpdir.tar;'
                ' /bin/rm $d/autopkgtest-tmpdir.tar; exit 0; '
                'fi; done; exit 1' % directories],
               downp=True, timeout=copy_timeout)
    adtlog.debug('cmd_reboot: restored downtmp after reboot')


def get_downtmp_host():
    '''Return host directory of the testbed's downtmp dir, if supported'''

    for cap in caller.hook_capabilities():
        if cap.startswith('downtmp-host='):
            return cap.split('=', 1)[1]
    return None


def copytree(src, dst):
    '''Like shutils.copytree(), but merges with existing dst'''

    if not os.path.exists(dst):
        shutil.copytree(src, dst, symlinks=True)
        return

    for f in os.listdir(src):
        fsrc = os.path.join(src, f)
        subprocess.check_call(['cp', '-r', '--preserve=timestamps,links',
                               '--target-directory', dst, fsrc])


def copyup_shareddir(tb, host, is_dir, downtmp_host):
    adtlog.debug('copyup_shareddir: tb %s host %s is_dir %s downtmp_host %s'
                 % (tb, host, is_dir, downtmp_host))

    host = os.path.normpath(host)
    tb = os.path.normpath(tb)
    downtmp_host = os.path.normpath(downtmp_host)

    timeout_start(copy_timeout)
    try:
        tb_tmp = None
        if tb.startswith(downtmp):
            # translate into host path
            tb = downtmp_host + tb[len(downtmp):]
        else:
            tb_tmp = os.path.join(downtmp, os.path.basename(host))
            adtlog.debug('copyup_shareddir: tb path %s is not already in '
                         'downtmp, copying to %s' % (tb, tb_tmp))
            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', tb, tb_tmp],
                       downp=True)
            # translate into host path
            tb = os.path.join(downtmp_host, os.path.basename(host))

        if tb == host:
            tb_tmp = None
        else:
            adtlog.debug('copyup_shareddir: tb(host) %s is not already at '
                         'destination %s, copying' % (tb, host))
            if is_dir:
                copytree(tb, host)
            else:
                shutil.copy(tb, host)

        if tb_tmp:
            adtlog.debug('copyup_shareddir: rm intermediate copy: %s' % tb)
            check_exec(['/bin/rm', '-rf', tb_tmp], downp=True)
    finally:
        timeout_stop()


def copydown_shareddir(host, tb, is_dir, downtmp_host):
    adtlog.debug('copydown_shareddir: host %s tb %s is_dir %s downtmp_host %s'
                 % (host, tb, is_dir, downtmp_host))

    host = os.path.normpath(host)
    tb = os.path.normpath(tb)
    downtmp_host = os.path.normpath(downtmp_host)

    timeout_start(copy_timeout)
    try:
        host_tmp = None
        if host.startswith(downtmp_host):
            # translate into tb path
            host = downtmp + host[len(downtmp_host):]
        else:
            host_tmp = os.path.join(downtmp_host, os.path.basename(tb))
            if is_dir:
                if os.path.exists(host_tmp):
                    try:
                        shutil.rmtree(host_tmp)
                    except OSError as e:
                        adtlog.warning('cannot remove old %s, moving it '
                                       'instead: %s' % (host_tmp, e))
                        # some undeletable files? hm, move it aside instead
                        counter = 0
                        while True:
                            p = host_tmp + '.old%i' % counter
                            if not os.path.exists(p):
                                os.rename(host_tmp, p)
                                break
                            counter += 1

                shutil.copytree(host, host_tmp, symlinks=True)
            else:
                shutil.copy(host, host_tmp)
            # translate into tb path
            host = os.path.join(downtmp, os.path.basename(tb))

        if host == tb:
            host_tmp = None
        else:
            check_exec(['/bin/rm', '-rf', tb], downp=True)
            check_exec(['/bin/cp', '-r', '--preserve=timestamps,links', host, tb],
                       downp=True)
        if host_tmp:
            (is_dir and shutil.rmtree or os.unlink)(host_tmp)
    finally:
        timeout_stop()


def copyupdown(c, ce, upp):
    cmdnumargs(c, ce, 2)
    copyupdown_internal(ce[0], c[1:], upp)


def copyupdown_internal(wh, sd, upp):
    '''Copy up/down a file or dir.

    wh: 'copyup' or 'copydown'
    sd: (source, destination) paths
    upp: True for copyup, False for copydown
    '''
    if not downtmp:
        bomb("%s when not open" % wh)
    if not sd[0] or not sd[1]:
        bomb("%s paths must be nonempty" % wh)
    dirsp = sd[0][-1] == '/'
    if dirsp != (sd[1][-1] == '/'):
        bomb("%s paths must agree about directoryness"
             " (presence or absence of trailing /)" % wh)

    # if we have a shared directory, we just need to copy it from/to there; in
    # most cases, it's testbed end is already in the downtmp dir
    downtmp_host = get_downtmp_host()
    if downtmp_host:
        try:
            if upp:
                copyup_shareddir(sd[0], sd[1], dirsp, downtmp_host)
            else:
                copydown_shareddir(sd[0], sd[1], dirsp, downtmp_host)
            return
        except Timeout:
            raise FailedCmd(['timeout'])
        except (shutil.Error, subprocess.CalledProcessError) as e:
            adtlog.debug('Cannot copy %s to %s through shared dir: %s, falling back to tar' %
                         (sd[0], sd[1], str(e)))

    isrc = 0
    idst = 1
    ilocal = 0 + upp
    iremote = 1 - upp

    deststdout = devnull_read
    srcstdin = devnull_read
    remfileq = shlex.quote(sd[iremote])
    if not dirsp:
        rune = 'cat %s%s' % ('><'[upp], remfileq)
        if upp:
            deststdout = open(sd[idst], 'wb')
        else:
            srcstdin = open(sd[isrc], 'rb')
            status = os.fstat(srcstdin.fileno())
            if status.st_mode & 0o111:
                rune += '; chmod +x -- %s' % (remfileq)
        localcmdl = ['cat']
    else:
        taropts = [None, None]
        taropts[isrc] = '--warning=none -c .'
        taropts[idst] = '--warning=none --preserve-permissions --extract ' \
                        '--no-same-owner'

        rune = 'cd %s; tar %s -f -' % (remfileq, taropts[iremote])
        if upp:
            try:
                os.mkdir(sd[ilocal])
            except (IOError, OSError) as oe:
                if oe.errno != errno.EEXIST:
                    raise
        else:
            rune = ('if ! test -d %s; then mkdir -- %s; fi; ' % (
                remfileq, remfileq)
            ) + rune

        localcmdl = ['tar', '--directory', sd[ilocal]] + (
            ('%s -f -' % taropts[ilocal]).split()
        )
    downcmdl = auxverb + ['sh', '-ec', rune]

    if upp:
        cmdls = (downcmdl, localcmdl)
    else:
        cmdls = (localcmdl, downcmdl)

    adtlog.debug(str(["cmdls", str(cmdls)]))
    adtlog.debug(str(["srcstdin", str(srcstdin), "deststdout",
                      str(deststdout), "devnull_read", devnull_read]))

    subprocs = [None, None]
    adtlog.debug(" +< %s" % ' '.join(cmdls[0]))
    subprocs[0] = subprocess.Popen(cmdls[0], stdin=srcstdin,
                                   stdout=subprocess.PIPE)
    adtlog.debug(" +> %s" % ' '.join(cmdls[1]))
    subprocs[1] = subprocess.Popen(cmdls[1], stdin=subprocs[0].stdout,
                                   stdout=deststdout)
    subprocs[0].stdout.close()
    try:
        timeout_start(copy_timeout)
        for sdn in [1, 0]:
            adtlog.debug(" +" + "<>"[sdn] + "?")
            status = subprocs[sdn].wait()
            if not (status == 0 or (sdn == 0 and status == -13)):
                timeout_stop()
                adtlog.info("%s %s failed, status %d" %
                            (wh, ['source', 'destination'][sdn], status))
                raise FailedCmd(['copy-failed'])
        timeout_stop()
    except Timeout:
        for sdn in [1, 0]:
            subprocs[sdn].kill()
            subprocs[sdn].wait()
        raise FailedCmd(['timeout'])


def cmd_copydown(c, ce):
    copyupdown(c, ce, False)


def cmd_copyup(c, ce):
    copyupdown(c, ce, True)


def cmd_shell(c, ce):
    cmdnumargs(c, ce, 1, None)
    if not downtmp:
        bomb("`shell' when not open")
    # runners can provide a hook if they need a special treatment
    try:
        caller.hook_shell(*c[1:])
    except AttributeError:
        adtlog.debug('cmd_shell: using default shell command, dir %s' % c[1])
        cmd = 'cd "%s"; ' % c[1]
        for e in c[2:]:
            cmd += 'export "%s"; ' % e
        # use host's $TERM to provide a sane shell
        try:
            cmd += 'export TERM="%s"; ' % os.environ['TERM']
        except KeyError:
            pass
        cmd += 'bash -i'
        try:
            with open('/dev/tty', 'rb') as sin:
                with open('/dev/tty', 'wb') as sout:
                    with open('/dev/tty', 'wb') as serr:
                        subprocess.call(auxverb + ['sh', '-c', cmd],
                                        stdin=sin, stdout=sout, stderr=serr)
        except (OSError, IOError) as e:
            adtlog.error('Cannot run shell: %s' % e)


def command():
    sys.stdout.flush()
    ce = sys.stdin.readline()
    if not ce:
        bomb('end of file - caller quit?')
    ce = ce.rstrip().split()
    c = list(map(url_unquote, ce))
    if not c:
        bomb('empty commands are not permitted')
    adtlog.debug('executing ' + ' '.join(ce))
    c_lookup = c[0].replace('-', '_')
    try:
        f = globals()['cmd_' + c_lookup]
    except KeyError:
        bomb("unknown command `%s'" % ce[0])
    try:
        r = f(c, ce)
        if not r:
            r = []
        r.insert(0, 'ok')
    except FailedCmd as fc:
        r = fc.e
    print(' '.join(r))


signal_list = [signal.SIGHUP, signal.SIGTERM,
               signal.SIGINT, signal.SIGPIPE]


def sethandlers(f):
    for signum in signal_list:
        signal.signal(signum, f)


def cleanup():
    global downtmp, cleaning
    adtlog.debug("cleanup...")
    sethandlers(signal.SIG_DFL)
    # avoid recursion if something bomb()s in hook_cleanup()
    if not cleaning:
        cleaning = True
        if downtmp:
            caller.hook_cleanup()
        cleaning = False
        downtmp = None


def error_cleanup():
    try:
        ok = False
        try:
            cleanup()
            ok = True
        except Quit as q:
            sys.stderr.write(q.m)
            sys.stderr.write('\n')
        except Exception:
            sys.stderr.write('Unexpected cleanup error:\n')
            traceback.print_exc()
            sys.stderr.write('\n')
        if not ok:
            sys.stderr.write('while cleaning up because of another error:\n')
    except Exception:
        pass


def prepare():
    def handler(sig, *any):
        cleanup()
        os.kill(os.getpid(), sig)
    sethandlers(handler)


def cmd_auxverb_debug_fail(c, ce):
    cmdnumargs(c, ce)
    try:
        adtlog.info(caller.hook_debug_fail())
    except AttributeError:
        pass


def mainloop():
    global in_mainloop

    if (fcntl.fcntl(0, fcntl.F_GETFL) & os.O_NONBLOCK) != 0:
        # Some versions of autopkgtest-virt-* would try to cope
        # by sleeping 0.1 when they got EOF/EAGAIN.
        # But that causes us to hang, looping, when our caller goes away.
        #
        # Even though this is an egregious violation of Unix calling
        # conventions, and would be entirely the caller's fault,
        # detecting this situation and crashing is sensible.
        # Otherwise we would produce misleading error messages -
        # stdin.readline() returns '', not EAGAIN, which looks just like EOF.
        # (And in theory, we might read, and act on, partial commands!)
        bomb('stdin to autopkgtest virt server is nonblocking!')

    in_mainloop = True

    try:
        while True:
            command()
    except Quit as q:
        error_cleanup()
        if q.m:
            sys.stderr.write(q.m)
            sys.stderr.write('\n')
        sys.exit(q.ec)
    except Exception:
        error_cleanup()
        sys.stderr.write('Unexpected error:\n')
        traceback.print_exc()
        sys.exit(16)
    finally:
        in_mainloop = False


def main():
    ok()
    prepare()
    mainloop()
